package org.iwsec.pws.pwscup2016.check;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.regex.Pattern;

import org.iwsec.pws.pwscup2016.Utility;

public class CheckerT {

	static final public int MAX_COLUMN_COUNT = 78;
	static final public int MAX_ROW_COUNT = 1000 * 1000;

	static final public List<? extends Pattern> PATTERNS;

	static {
		try {
			PATTERNS = Collections.unmodifiableList(CheckUtility.readPatterns("patternsT.txt"));
		}
		catch (IOException e) {
			throw new ExceptionInInitializerError(e);
		}
	}

	static public ArrayList<Predicate<CharSequence>> newColumnPredicates() {
		ArrayList<Predicate<CharSequence>> columnPredicates = new ArrayList<>(PATTERNS.size());
		for (Pattern p : PATTERNS) {
			columnPredicates.add(s -> p.matcher(s).matches());
		}
		return columnPredicates;
	}

	static public Function<Integer, String> newRowCheck() {
		return r -> (r <= 0 ? "not enough rows" : null);
	}

	static public void main(String... args) throws IOException {
		Collection<? extends Message> messages = newMessages(args);
		if (messages.isEmpty()) {
			return;
		}
		for (Message m : messages) {
			System.err.println(m);
		}
		System.exit(1);
	}

	static public Collection<? extends Message> newMessages(String... args) throws IOException {
		if (args.length < 1) {
			throw new IllegalArgumentException(String.format("A transaction data file path must be specified as the argument."));
		}
		Path path = Paths.get(args[0]);

		return newMessages(path);
	}

	static public Collection<? extends Message> newMessages(Path path) throws IOException {
		try (BufferedReader reader = Files.newBufferedReader(path, StandardCharsets.ISO_8859_1)) {
			return newMessages(reader);
		}
	}

	static public Collection<? extends Message> newMessages(BufferedReader reader) throws IOException {
		ReaderWithCheck r = new ReaderWithCheck(reader, MAX_COLUMN_COUNT, MAX_ROW_COUNT);
		return new CheckerT().newMessages(newColumnPredicates(), newRowCheck(), r);
	}

	public Collection<? extends Message> newMessages(Collection<? extends Predicate<? super String>> columnPredicates, Function<? super Integer, ? extends String> rowCheck, ReaderWithCheck reader) throws IOException {
		List<Message> messages = new ArrayList<>();
		String line;
		try {
			while ((line = reader.readLine()) != null) {
				Message message = getMessage(columnPredicates, reader.getLineNumber(), line);
				if (message != null) {
					messages.add(message);
				}
			}
			int lineNumber = reader.getLineNumber();
			String s = rowCheck.apply(lineNumber);
			if (s != null) {
				Message message = new Message(lineNumber <= 0 ? 1 : lineNumber, s);
				messages.add(message);
			}
		}
		catch (CheckException e) {
			messages.add(new Message(reader.getLineNumber(), e.getMessage()));
		}
		return messages;
	}

	protected Message getMessage(Collection<? extends Predicate<? super String>> columnPredicates, int lineNumber, String line) {
		List<String> cellList = Utility.separate(line, ",");
		if (cellList.size() != columnPredicates.size()) {
			String prefix = (cellList.size() < columnPredicates.size() ? "not enough" : "too many");
			String message = String.format("%s cells (expected %d, but actually %d)", prefix, columnPredicates.size(), cellList.size());
			return new Message(lineNumber, message);
		}
		Iterator<String> cellIterator = cellList.iterator();
		Iterator<? extends Predicate<? super String>> predicateIterator = columnPredicates.iterator();
		for (int i = 0; i < cellList.size(); i++) {
			String cell = cellIterator.next();
			Predicate<? super String> p = predicateIterator.next();
			if (!p.test(cell)) {
				String message = String.format("column %d: invalid value (%s)", i + 1, cell);
				return new Message(lineNumber, message);
			}
		}
		return null;
	}
}
