package org.iwsec.pws.pwscup2016.ut;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

import org.iwsec.pws.pwscup2016.ExampleUtility;
import org.iwsec.pws.pwscup2016.Utility;

public class RfmT {
	static protected class Rfm {
		private long rInDate;
		private final Set<String> fInInvoice;
		private long mInPenny;

		public Rfm() {
			rInDate = Long.MIN_VALUE;
			fInInvoice = new HashSet<>();
		}

		public long getRInDate() {
			return rInDate;
		}

		public int getFInCount() {
			return fInInvoice.size();
		}

		public long getMInPenny() {
			return mInPenny;
		}

		public void update(long date, String invoice, long penny) {
			if (date > rInDate) {
				rInDate = date;
			}
			fInInvoice.add(invoice);
			mInPenny += penny;
		}

		@Override
		public String toString() {
			return String.format("r=%s, f=%s, m=%s", getRInDate(), getFInCount(), getMInPenny());
		}
	}

	static protected class RfmRank {
		static final public RfmRank MAX = new RfmRank(9, 9, 9);

		static final public List<? extends Long> R_THRESHOLDS;

		static {
			List<String> dateList = new ArrayList<>();
			dateList.add("2011/03/27");
			dateList.add("2011/06/19");
			dateList.add("2011/08/28");
			dateList.add("2011/10/02");
			dateList.add("2011/10/23");
			dateList.add("2011/11/13");
			dateList.add("2011/11/20");
			dateList.add("2011/11/27");
			dateList.add("2011/12/04"); // original: dateList.add("2011/12/11");
			assert dateList.size() == MAX.getR();
			R_THRESHOLDS = Collections.unmodifiableList(dateList.stream().map(Utility::parseDate).collect(Collectors.toList()));
		}

		static final public List<? extends Integer> F_THRESHOLDS;

		static {
			List<Integer> countList = new ArrayList<>();
			countList.add(2);
			countList.add(3); // original: countList.add(2);
			countList.add(4); // original: countList.add(2);
			countList.add(5); // original: countList.add(3);
			countList.add(6); // original: countList.add(3);
			countList.add(7); // original: countList.add(4);
			countList.add(8); // original: countList.add(5);
			countList.add(9); // original: countList.add(7);
			countList.add(10);
			assert countList.size() == MAX.getF();
			F_THRESHOLDS = Collections.unmodifiableList(countList);
		}

		static final public List<? extends Long> M_THRESHOLDS;

		static {
			List<Long> pennyList = new ArrayList<>();
			pennyList.add(200 * 100L);
			pennyList.add(300 * 100L);
			pennyList.add(400 * 100L);
			pennyList.add(500 * 100L);
			pennyList.add(700 * 100L);
			pennyList.add(1000 * 100L);
			pennyList.add(1400 * 100L);
			pennyList.add(2100 * 100L);
			pennyList.add(3700 * 100L);
			assert pennyList.size() == MAX.getM();
			M_THRESHOLDS = Collections.unmodifiableList(pennyList);
		}

		static public RfmRank newRank(Rfm rfm) {
			int r = getRankOfR(rfm);
			int f = getRankOfF(rfm);
			int m = getRankOfM(rfm);
			return new RfmRank(r, f, m);
		}

		static protected <T extends Comparable<T>> int getRank(List<? extends T> thresholds, T value) {
			int i = 0;
			for (T t : thresholds) {
				if (value.compareTo(t) < 0) {
					return i;
				}
				i++;
			}
			return i;
		}

		static public int getRankOfR(Rfm rfm) {
			return getRank(R_THRESHOLDS, rfm.getRInDate());
		}

		static public int getRankOfF(Rfm rfm) {
			return getRank(F_THRESHOLDS, rfm.getFInCount());
		}

		static public int getRankOfM(Rfm rfm) {
			long penny = rfm.getMInPenny();
			return getRank(M_THRESHOLDS, penny);
		}

		private final List<? extends Integer> rfm;

		public RfmRank(int r, int f, int m) {
			rfm = Arrays.asList(r, f, m);

			if (MAX == null) {
				return;
			}

			if (r < 0 || r > MAX.getR()) {
				throw new IllegalArgumentException(String.format("r must be in range [0, %d]: %s", MAX.getR(), r));
			}
			if (f < 0 || f > MAX.getF()) {
				throw new IllegalArgumentException(String.format("f must be in range [0, %d]: %s", MAX.getF(), f));
			}
			if (m < 0 || m > MAX.getM()) {
				throw new IllegalArgumentException(String.format("m must be in range [0, %d]: %s", MAX.getM(), m));
			}
		}

		public List<? extends Integer> asList() {
			return rfm;
		}

		public int getR() {
			return asList().get(0);
		}

		public int getF() {
			return asList().get(1);
		}

		public int getM() {
			return asList().get(2);
		}

		@Override
		public int hashCode() {
			return asList().hashCode();
		}

		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof RfmRank)) {
				return false;
			}
			RfmRank that = (RfmRank) obj;
			return asList().equals(that.asList());
		}

		@Override
		public String toString() {
			return asList().toString();
		}
	}

	protected double calculate(List<? extends List<? extends String>> mTable, List<? extends List<? extends String>> tTable, List<? extends List<? extends String>> m2Table, List<? extends List<? extends String>> t2Table) {
		Map<? extends RfmRank, ? extends Integer> rfmCountMap = newRankCountMap(mTable, tTable);
		Map<? extends RfmRank, ? extends Integer> rfmCount2Map = newRankCountMap(m2Table, t2Table);

		if (!rfmCountMap.keySet().equals(rfmCount2Map.keySet())) {
			throw new AssertionError();
		}

		double rmse = 0;
		for (Entry<? extends RfmRank, ? extends Integer> entry : rfmCountMap.entrySet()) {
			int count = entry.getValue();
			int count2 = rfmCount2Map.get(entry.getKey());
			rmse += Math.pow(count - count2, 2);
		}
		rmse /= rfmCountMap.size();
		rmse = Math.sqrt(rmse);

		// l傫ǂƂ̂ŁAlɂB
		return -rmse;
	}

	protected Map<? extends RfmRank, ? extends Integer> newRankCountMap(List<? extends List<? extends String>> mTable, List<? extends List<? extends String>> tTable) {
		Map<RfmRank, Integer> rankCountMap = new HashMap<>();
		for (int r = 0; r <= RfmRank.MAX.getR(); r++) {
			for (int f = 0; f <= RfmRank.MAX.getF(); f++) {
				for (int m = 0; m <= RfmRank.MAX.getM(); m++) {
					rankCountMap.put(new RfmRank(r, f, m), 0);
				}
			}
		}

		Map<? extends String, ? extends Rfm> customerRfmMap = newCustomerRfmMap(mTable, tTable);
		customerRfmMap.forEach((customer, rfm) -> {
			RfmRank rank = RfmRank.newRank(rfm);
			int count = rankCountMap.get(rank);
			rankCountMap.put(rank, count + 1);
		});

		return rankCountMap;
	}

	protected Map<? extends String, ? extends Rfm> newCustomerRfmMap(List<? extends List<? extends String>> mTable, List<? extends List<? extends String>> tTable) {
		Map<String, Rfm> customerRfm = new HashMap<>();

		HashSet<? extends String> customerSet = new HashSet<>(ExampleUtility.getList(mTable, ExampleUtility.INDEX_M_CUSTOMER));

		for (List<? extends String> row : tTable) {
			String customer = row.get(ExampleUtility.INDEX_T_CUSTOMER);
			if (!customerSet.contains(customer)) {
				continue;
			}
			Rfm rfm = customerRfm.get(customer);
			if (rfm == null) {
				customerRfm.put(customer, rfm = new Rfm());
			}
			long date = ExampleUtility.parseDate(row.get(ExampleUtility.INDEX_T_DATE_YMD), row.get(ExampleUtility.INDEX_T_DATE_HM));
			String invoice = row.get(ExampleUtility.INDEX_T_INVOICE);
			long penny = ExampleUtility.parsePriceInPenny(row.get(ExampleUtility.INDEX_T_PRICE));
			penny *= Integer.parseInt(row.get(ExampleUtility.INDEX_T_QUANTITY));
			rfm.update(date, invoice, penny);
		}
		return customerRfm;
	}

	static public void main(String... args) throws IOException {
		Path in = null;
		Path out = null;
		if (args.length > 0) {
			in = Paths.get(args[0]);
			if (args.length > 1) {
				out = Paths.get(args[1]);
			}
		}

		ArrayList<ArrayList<ArrayList<String>>> tableList = ExampleUtility.read(in);

		RfmT one = new RfmT();
		double value = one.calculate(tableList.get(0), tableList.get(1), tableList.get(2), tableList.get(3));
		tableList.clear();
		tableList.add(ExampleUtility.asTable(Arrays.asList(value), v -> String.format("%f", v)));

		ExampleUtility.write(out, tableList);
	}
}
